from sys import byteorder
from bitstring import BitArray, Bits, ConstBitStream
from .varint_codec import VarintCodec
from .field_codec import FieldCodec

def to_sixbit_ascii(character: str):
    upper_character = character.upper()
    ascii_code = int(upper_character.encode('ascii')[0])

    if ascii_code >= 0x40 and ascii_code <= 0x5f:
        sixbit_code = ascii_code - 0x40
    elif ascii_code >= 0x20 and ascii_code <= 0x3f:
        sixbit_code = ascii_code
    else:
        sixbit_code = 0x3f  # out of bounds values are encoded as '?'

    return sixbit_code


def from_sixbit_ascii(sixbit_code: int):
    if sixbit_code <= 0x1f:
        ascii_code = sixbit_code + 0x40
    # Other characters map directly
    else:
        ascii_code = sixbit_code

    return bytes([ascii_code]).decode('ascii')


class AsciiStringCodec(FieldCodec):
    def __init__(self, max_length: int = 128, bits_per_char: int = 7, tail=False, **kwargs):
        self.max_length = int(max_length)
        self.bits_per_char = bits_per_char
        self.tail = tail
        self.string_len_codec = VarintCodec(min_value=0, max_value=max_length)

    def encode(self, value: str):
        if not self.tail:
            value = value[0:self.max_length]
        else:
            value = value[-self.max_length:]
        length_bits, _ = self.string_len_codec.encode(len(value))
        encoded_bits = BitArray()
        encoded_bits.append(length_bits)

        string_bytes = value.encode('ascii')
        if self.bits_per_char == 7:
            compressed_bytes = bytearray()
            for sb in string_bytes:
                if sb > 0x7f:
                    # Replace out of bounds values with "?"
                    sb = 0x3f
                compressed_bytes.append(sb)
                encoded_bits.append(Bits(bytes=[sb]))
        elif self.bits_per_char == 6:
            compressed_bytes = bytearray()
            for sb in string_bytes:
                sixbit_value = to_sixbit_ascii(sb)
                compressed_byte = from_sixbit_ascii(sixbit_value).encode('ascii')
                compressed_bytes.extend(compressed_byte)
                encoded_bits.append(Bits(bytes=[sixbit_value]), length=6)
        else:
            encoded_bits.append(Bits(bytes=value))
            compressed_bytes = string_bytes

        compressed_string = compressed_bytes.decode('ascii')
        return encoded_bits, compressed_string

    def decode(self, encoded_bits: ConstBitStream):
        counter = 2
        num_chars = self.string_len_codec.decode(encoded_bits)
        test = bytearray()
        for i in range(num_chars):
                test.append((encoded_bits.read('uint:7')))
        print(test.decode('ascii'))
        encoded_bitArray = encoded_bits.tobytes()
        returnBytes = bytearray(encoded_bitArray)
        truncatedBytes = bytearray()

        while counter <= num_chars + 1:
            truncatedBytes.append(returnBytes[counter])
            counter+=1
        
        return truncatedBytes.decode('ascii')

    @property
    def max_length_bits(self):
        return self.string_len_codec.max_length_bits + (self.max_length * self.bits_per_char)

    @property
    def min_length_bits(self):
        return self.string_len_codec.max_length_bits


import re
from bitstring import ConstBitStream, BitArray
from .field_codec import FieldCodec
import field_codecs as field_codecs
from rospy import AnyMsg
from typing import Union
import roslib.message


class RosMsgFieldCodec(FieldCodec):
    def __init__(self, ros_type: Union[str, AnyMsg], fields: dict = None, **kwargs):
        self.fields = fields
        self.ros_type = ros_type

        if isinstance(ros_type, str):
            msg_class = roslib.message.get_message_class(ros_type)
            if not msg_class:
                raise TypeError('Invalid ROS type "{}" for message codec {}'.format(ros_type, id))
        else:
            msg_class = ros_type
        self.ros_msg_class = msg_class

        self.field_codecs = {}
        # Handle the case where we have to infer the field list from the ROS msg, since the user didn't specify one.
        if not self.fields:
            slot_names = msg_class.__slots__
            slot_types = msg_class._slot_types

            slots = dict(zip(slot_names, slot_types))

            new_fields = dict.fromkeys(slot_names)
            for field_name in new_fields:
                msg_type = slots[field_name]
                if msg_type in field_codecs.field_codec_classes:
                    field_params = {'codec': msg_type}
                else:
                    # There are two cases here: arrays or nested ROS messages
                    if msg_type.endswith(']'):
                        # either a variable or fixed length array
                        matched = re.match(r'(?P<element_type>.*)\[(?P<array_len>\d*)\]', msg_type)
                        if matched['array_len']:
                            array_len = int(matched['array_len'])
                            field_params = {'codec': 'fixed_len_array',
                                            'length': array_len}
                        else:
                            field_params = {'codec': 'variable_len_array',
                                            'max_length': 10}  # TODO: maybe take this as a parameter?
                        # and the element type is either a standard type or a ROS message
                        if matched['element_type'] in field_codecs.field_codec_classes:
                            field_params['element_type'] = matched['element_type']
                        else:
                            field_params['element_type'] = 'msg'
                            field_params['element_params'] = {'ros_type': matched['element_type']}
                    else:
                        # It's a nested ROS message
                        field_params = {'codec': 'msg',
                                        'ros_type': msg_type}

                new_fields[field_name] = field_params

            self.fields = new_fields

        # Now, build the list of codec classes for each field
        self.field_codecs = {}
        for field_name, field_params in self.fields.items():
            # print(field_name, field_params['codec'])
            # print(field_codecs.field_codec_classes[field_params['codec']])
            try:
                if field_params['codec'] not in field_codecs.metadata_decoders.keys():
                    self.field_codecs[field_name] = field_codecs.field_codec_classes[field_params['codec']](**field_params)
                else:
                    self.field_codecs[field_name] = None
            except KeyError as e:
                raise KeyError("Error parsing codec config for {}.  Got params:\n{}\nError: {}".format(field_name,
                                                                                                       field_params,
                                                                                                       e))

    def encode(self, message: AnyMsg):
        # ROS messages use __slots__, so we can't use __dict__ or vars() to get the attributes as a dict
        message_dict = {}
        for field in message.__slots__:
            message_dict[field] = getattr(message, field)

        encoded_bits = BitArray()
        encoded_dict = {}
        for field_name, field_params in self.fields.items():
            try:
                field_codec = self.field_codecs[field_name]
                if not field_codec:
                    continue
                field_bits, encoded_dict[field_name] = field_codec.encode(message_dict[field_name])
                encoded_bits.append(field_bits)
            except Exception as e:
                #print("Codec: {}, max len bits {}".format(field_codec, field_codec.max_length_bits))
                raise Exception('Error encoding field "{}" with codec {} (max len bits {})'.format(field_name, field_codec, field_codec.max_length_bits)).with_traceback(e.__traceback__)
        return encoded_bits, encoded_dict

    def decode(self, bits_to_decode: ConstBitStream):
        # We go through the bits in sequence until we are decoded.
        # The ConstBitStream has an internal read pointer.
        decoded_message = {}
        for field_name, field_params in self.fields.items():
            field_codec = self.field_codecs[field_name]
            if not field_codec:
                continue
            ## print("Ros decode field {}".format(field_name))
            decoded_message[field_name] = field_codec.decode(bits_to_decode)
        ## print("Ros decode got {}".format(decoded_message))
        return self.ros_msg_class(**decoded_message)

    def decode_as_dict(self, bits_to_decode: ConstBitStream):
        # This function is used to generate an object that we can use for calculating CRCs in the message codec
        # It seems like a bit of a hack, but the alternative would require adding a second return value to every field
        # decoder.  Since this only affects ROS messages, this keeps the bloat down.
        decoded_message = {}
        for field_name, field_params in self.fields.items():
            field_codec = self.field_codecs[field_name]
            if hasattr(field_codec, 'decode_as_dict'):
                decoded_message[field_name] = field_codec.decode_as_dict(bits_to_decode)
            else:
                decoded_message[field_name] = field_codec.decode(bits_to_decode)
        ## print("Ros decode got {}".format(decoded_message))
        return decoded_message

    @property
    def min_length_bits(self):
        return sum([c.min_length_bits for c in self.field_codecs.values()])

    @property
    def max_length_bits(self):
        return sum([c.max_length_bits for c in self.field_codecs.values()])




from bitstring import ConstBitStream, Bits
from .field_codec import FieldCodec
from .varint_codec import VarintCodec
from math import ceil, log2


class FixedIntCodec(VarintCodec):
    def __init__(self, num_bits: int, signed: bool = False, min_value: int = None, resolution: int = 1,
                 little_endian: bool = False, **kwargs):
        self.num_bits = num_bits
        if min_value:
            self.min_value = int(min_value)
        else:
            if signed:
                self.min_value = -1 * resolution * 2**(self.num_bits - 1)
            else:
                self.min_value = 0
        self.max_value = self.min_value + (2**self.num_bits * resolution)
        self.resolution = int(resolution)
        self.signed = signed
        self.little_endian = little_endian




class Int8Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(Int8Codec, self).__init__(num_bits=8, signed=True)


class Int16Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(Int16Codec, self).__init__(num_bits=16, signed=True)


class Int32Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(Int32Codec, self).__init__(num_bits=32, signed=True)


class Int64Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(Int64Codec, self).__init__(num_bits=64, signed=True)


class UInt8Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(UInt8Codec, self).__init__(num_bits=8, signed=False)


class UInt16Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(UInt16Codec, self).__init__(num_bits=16, signed=False)


class UInt32Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(UInt32Codec, self).__init__(num_bits=32, signed=False)


class UInt64Codec(FixedIntCodec):
    def __init__(self, **kwargs):
        super(UInt32Codec, self).__init__(num_bits=64, signed=False)


from bitstring import ConstBitStream, Bits
from .field_codec import FieldCodec
from math import ceil, log2


class CclLatLonCodec(FieldCodec):
    def __init__(self, **kwargs):
        pass

    def encode(self, value: float):
        encoded_value = int(value * ((2 ** 23 - 1) / 180.0))
        encoded_bits = Bits(intle=encoded_value, length=24)

        return encoded_bits, encoded_value

    def decode(self, bits_to_decode: ConstBitStream):
        scale = bits_to_decode.read('intle:24')
        value = scale * (180.0 / (2**23 - 1))
        return value

    @property
    def max_length_bits(self):
        return 24

    @property
    def min_length_bits(self):
        return 24


from bitstring import BitArray, ConstBitStream
from .field_codec import FieldCodec
from .varint_codec import VarintCodec


class BytesCodec(FieldCodec):
    def __init__(self, max_length: int, **kwargs):
        self.max_length = max_length
        self.length_codec = VarintCodec(min_value=0, max_value=self.max_length)

    def encode(self, value: bytes):
        value = value[0:self.max_length]
        length_bits, _ = self.length_codec.encode(len(value))
        value_bits = BitArray(bytes=value)
        value_bits.prepend(length_bits)
        return value_bits, value

    def decode(self, bits_to_decode: ConstBitStream):
        num_bytes = self.length_codec.decode(bits_to_decode)
        value = bits_to_decode.read('bytes:' + str(num_bytes))
        return value

    def min_length_bits(self):
        return self.length_codec.max_length_bits

    def max_length_bits(self):
        return self.length_codec.max_length_bits + (8 * self.max_length)


from bitstring import ConstBitStream, Bits
from .field_codec import FieldCodec
from math import ceil, log2


class VarintCodec(FieldCodec):
    def __init__(self, min_value: int, max_value: int, resolution: int = 1, little_endian: bool = False, **kwargs):
        self.max_value = int(max_value)
        self.min_value = int(min_value)
        self.resolution = int(resolution)

        self.value_range = max_value - min_value
        num_values = (self.value_range // self.resolution) + 1
        self.num_bits = ceil(log2(num_values))
        self.little_endian = little_endian

    def encode(self, value: int):
        value = int(value)
        if value < self.min_value:
            value = self.min_value
        elif value > self.max_value:
            value = self.max_value
        offset = value - self.min_value
        discretized_offset = offset // self.resolution
        encoded_value = self.min_value + (discretized_offset * self.resolution)
        if self.little_endian:
            encoded_bits = Bits(uintle=discretized_offset, length=self.num_bits)
        else:
            encoded_bits = Bits(uint=discretized_offset, length=self.num_bits)
        ## print("encode varint {} bits as {}".format(num_bits, encoded_bits))
        return encoded_bits, encoded_value

    def decode(self, bits_to_decode: ConstBitStream):
        if self.little_endian:
            discretized_offset = bits_to_decode.read('uintle:{}'.format(self.num_bits))
        else:
            discretized_offset = bits_to_decode.read('uint:{}'.format(self.num_bits))

        value = self.min_value + (discretized_offset * self.resolution)
        return value

    @property
    def max_length_bits(self):
        return self.num_bits

    @property
    def min_length_bits(self):
        return self.num_bits

    def __repr__(self):
        return "VarintCodec {:x}: min_value={}, max_value={}, resolution={}".format(id(self),
                                                                                    self.min_value,
                                                                                    self.max_value,
                                                                                    self.resolution)


from bitstring import BitArray, ConstBitStream
from .field_codec import FieldCodec
import field_codecs as field_codecs
from typing import List


class FixedLenArrayCodec(FieldCodec):
    def __init__(self, element_type: str, length: int, element_params=None, **kwargs):
        self.length = length
        if element_params:
            self.element_field_codec = field_codecs.field_codec_classes[element_type](**element_params)
        else:
            self.element_field_codec = field_codecs.field_codec_classes[element_type]()

    def encode(self, value: List):
        value = value[0:self.length]
        value_bits = BitArray()
        encoded_value_list = []
        for element in value:
            element_bits, element_value = self.element_field_codec.encode(element)
            value_bits.append(element_bits)
            encoded_value_list.append(element_value)
        return value_bits, encoded_value_list

    def decode(self, bits_to_decode: ConstBitStream):
        decoded_list = []
        for i in range(self.length):
            element = self.element_field_codec.decode(bits_to_decode)
            decoded_list.append(element)
        return decoded_list

    @property
    def min_length_bits(self):
        return self.length * self.element_field_codec.max_length_bits

    @property
    def max_length_bits(self):
        return self.length * self.element_field_codec.max_length_bits


from .varint_codec import VarintCodec
from .fixedint_codec import FixedIntCodec
from .fixedint_codec import Int8Codec, Int16Codec, Int32Codec, Int64Codec
from .fixedint_codec import UInt8Codec, UInt16Codec, UInt32Codec, UInt64Codec
from .float_codec import FloatCodec
from .linspace_float_codec import LinspaceFloatCodec
from .ieee_float_codec import IeeeFloat32Codec, IeeeFloat64Codec
from .bool_codec import BoolCodec
from .string_codecs import AsciiStringCodec
from .bytes_codec import BytesCodec
from .variable_len_array_codec import VariableLenArrayCodec
from .fixed_len_array_codec import FixedLenArrayCodec
from .ros_msg_field_codec import RosMsgFieldCodec
from .ccl_latlon_codec import CclLatLonCodec
from .ccl_latlon_bcd_codec import CclLatLonBcdCodec
from .padding_codec import PaddingCodec
from .rostime_codec import RosTimeCodec
from .ros_message_codec import RosMessageCodec


field_codec_classes = {'integer': VarintCodec,
                       'fixedint': FixedIntCodec,
                       'varint': VarintCodec,
                       'float': FloatCodec,
                       'linspace_float': LinspaceFloatCodec,
                       'linspace': LinspaceFloatCodec,
                       'uint8': UInt8Codec,
                       'uint16': UInt16Codec,
                       'uint32': UInt32Codec,
                       'uint64': UInt64Codec,
                       'int8': Int8Codec,
                       'int16': Int16Codec,
                       'int32': Int32Codec,
                       'int64': Int64Codec,
                       'float32': IeeeFloat32Codec,
                       'float64': IeeeFloat64Codec,
                       'bool': BoolCodec,
                       'string': AsciiStringCodec,
                       'ascii': AsciiStringCodec,
                       'bytes': BytesCodec,
                       'msg': RosMsgFieldCodec,
                       'ros_msg': RosMsgFieldCodec,
                       'variable_len_array': VariableLenArrayCodec,
                       'fixed_len_array': FixedLenArrayCodec,
                       'ccl_latlon': CclLatLonCodec,
                       'ccl_latlon_bcd': CclLatLonBcdCodec,
                       'pad': PaddingCodec,
                       'padding': PaddingCodec,
                       'time': RosTimeCodec,
                       'rostime': RosTimeCodec,
                       }

metadata_decoders = {'src': 'packet.src',
                     'dest': 'packet.dest'}


metadata_encoders = {'dest': 'packet.dest'}




from abc import ABC, abstractmethod

class FieldCodec(ABC):
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def encode(self):
        pass

    @abstractmethod
    def decode(self):
        pass

    @property
    @abstractmethod
    def max_length_bits(self):
        pass

    @property
    @abstractmethod
    def min_length_bits(self):
        pass


from bitstring import BitArray, ConstBitStream
from .field_codec import FieldCodec
from .varint_codec import VarintCodec
import field_codecs as field_codecs
from typing import List


class VariableLenArrayCodec(FieldCodec):
    def __init__(self, element_type: str, max_length: int, element_params=None, **kwargs):
        self.max_length = max_length
        self.length_codec = VarintCodec(min_value=0, max_value=self.max_length)
        print('element_type', element_type)
        if element_params:
            self.element_field_codec = field_codecs.field_codec_classes[element_type](**element_params)
        else:
            self.element_field_codec = field_codecs.field_codec_classes[element_type]()

    def encode(self, value: List):
        value = value[0:self.max_length]
        length_bits, _ = self.length_codec.encode(len(value))
        value_bits = BitArray(length_bits)
        encoded_value_list = []
        for element in value:
            element_bits, element_value = self.element_field_codec.encode(element)
            value_bits.append(element_bits)
            encoded_value_list.append(element_value)
        return value_bits, encoded_value_list

    def decode(self, bits_to_decode: ConstBitStream):
        num_elements = self.length_codec.decode(bits_to_decode)
        decoded_list = []
        for i in range(num_elements):
            element = self.element_field_codec.decode(bits_to_decode)
            decoded_list.append(element)
        return decoded_list

    @property
    def min_length_bits(self):
        return self.length_codec.max_length_bits

    @property
    def max_length_bits(self):
        return self.length_codec.max_length_bits + (self.max_length * self.element_field_codec.max_length_bits)


from bitstring import BitArray, ConstBitStream
from .field_codec import FieldCodec


class BoolCodec(FieldCodec):
    def __init__(self, **kwargs):
        pass

    def encode(self, value: bool):
        value = bool(value)
        value_bits = BitArray(bool=value)
        return value_bits, value

    def decode(self, bits_to_decode: ConstBitStream):
        value = bits_to_decode.read('bool')
        return value

    def min_length_bits(self):
        return 1

    def max_length_bits(self):
        return 1


from bitstring import ConstBitStream, Bits
from .field_codec import FieldCodec
from math import modf, copysign


class CclLatLonBcdCodec(FieldCodec):
    ''' This codec is horrifically inefficient, but included for compatibility with messages like MDAT_RANGER '''
    def __init__(self, lat_not_lon=True, **kwargs):
        self.lat_not_lon = lat_not_lon
        pass

    def encode(self, value: float):
        sign = copysign(1, value)
        abs_value = abs(value)
        frac, degrees = modf(abs_value)
        degrees = int(degrees)
        minutes = frac * 60
        dec_min = int(minutes * 10000)

        if self.lat_not_lon:
            sign_char = 'a' if sign > 0 else 'c'
        else:
            sign_char = 'b' if sign > 0 else 'd'

        hex_string = "{:03d}{}{:06d}".format(degrees, sign_char, dec_min)
        encoded_value = sign * (degrees + (dec_min / 10000) / 60)
        encoded_bits = Bits(hex=hex_string)

        return encoded_bits, encoded_value

    def decode(self, bits_to_decode: ConstBitStream):
        string_bytes = bits_to_decode.read('bytes:5')
        hex_string = ''.join('{:02x}'.format(x) for x in string_bytes)

        degrees = int(hex_string[0:3])
        direction = 1 if hex_string[3] in ('a', 'b') else -1
        minutes = float(hex_string[4:10]) / 10000.

        value = direction * (degrees + (minutes / 60))
        return value

    @property
    def max_length_bits(self):
        return 40

    @property
    def min_length_bits(self):
        return 40


import field_codecs
import msgpack
from crccheck.crc import Crc8SaeJ1850, Crc32Iscsi
from bitstring import Bits, ConstBitStream
from copy import copy
import operator
from std_msgs.msg import Header
import rospy
class RosMessageCodec(object):
    def __init__(self, ros_type, fields_dict: dict = None, checksum=None):
        self.ros_type = ros_type
        self.packet_codec = None
        self.checksum = checksum
        
        if fields_dict:
            self.metadata_fields = []
            for field_name, field_params in fields_dict.items():
                # We only support a few metadata fields for encoding
                codec_name = field_params.get('codec', 'auto')
                if codec_name in set(list(field_codecs.metadata_decoders.keys()) + list(field_codecs.metadata_encoders.keys())):
                    self.metadata_fields.append({field_name: codec_name})
        else:
            self.metadata_fields = None

        self.root_field_codec = field_codecs.RosMsgFieldCodec(ros_type=ros_type, fields=fields_dict)
        
    def encode(self, ros_msg):
        encoded_bits, encoded_dict = self.root_field_codec.encode(ros_msg)

        if self.checksum:
            msgpack_bytes = msgpack.packb(encoded_dict)
            if self.checksum == 'crc8':
                calculated_crc = Crc8SaeJ1850.calc(msgpack_bytes)
                encoded_bits.append(Bits(uint=calculated_crc, length=8))
            elif self.checksum == 'crc32':
                calculated_crc = Crc32Iscsi.calc(msgpack_bytes)
                encoded_bits.append(Bits(uint=calculated_crc, length=32))
            #print("Encoded CRC: {}".format(hex(calculated_crc)))

        metadata_dict = self._encode_metadata(ros_msg)
        return encoded_bits, metadata_dict

    def decode(self, bits_to_decode: ConstBitStream, received_packet = None):
        rospy.loginfo("Decoding ROS message {}".format(self.ros_type))
        # Now, check CRC, if required.
        if self.checksum:
            # We need to decode this as a dict separately, so we need a new copy of the received bitstream.
            bits_copy = copy(bits_to_decode)
            bits_copy.pos = bits_to_decode.pos
            decoded_dict = self.root_field_codec.decode_as_dict(bits_copy)

        ros_msg = self.root_field_codec.decode(bits_to_decode)

        if self.checksum:
            msgpack_bytes = msgpack.packb(decoded_dict)
            # print("Msgpack length: {} bytes".format(len(msgpack_bytes)))
            if self.checksum == 'crc8':
                calculated_crc = Crc8SaeJ1850.calc(msgpack_bytes)
                # Next 8 bits of message are CRC
                received_crc = bits_to_decode.read('uint:8')
            elif self.checksum == 'crc32':
                calculated_crc = Crc32Iscsi.calc(msgpack_bytes)
                received_crc = bits_to_decode.read('uint:32')

            #print("Received CRC: {}".format(hex(received_crc)))
            if calculated_crc != received_crc:
                raise ValueError("Message CRC Mismatch")

        # Append metadata
        if received_packet:
            self._decode_metadata(ros_msg, received_packet)

        rospy.loginfo("ROS Message: {}".format(ros_msg))
        return ros_msg

    def _decode_metadata(self, ros_msg, received_packet = None):
        # This will decode metadata fields into the ros msg.
        # it only works on fields of te base ROS message, not any nested ROS messages (for now)
        if self.metadata_fields:
            for field_name, codec in self.metadata_fields.items():
                metadata_attribute = field_codecs.metadata_decoders[codec]
                value = operator.attrgetter(metadata_attribute)(received_packet)
                setattr(ros_msg, field_name, value)

    def _encode_metadata(self, ros_msg):
        # TODO
        return None

    @property
    def max_length_bits(self):
        if self.checksum == 'crc8':
            checksum_len = 8
        elif self.checksum == 'crc32':
            checksum_len = 32
        else:
            checksum_len = 0
        return self.root_field_codec.max_length_bits + checksum_len

    @property
    def min_length_bits(self):
        if self.checksum == 'crc8':
            checksum_len = 8
        elif self.checksum == 'crc32':
            checksum_len = 32
        else:
            checksum_len = 0
        return self.root_field_codec.min_length_bits + checksum_len



from bitstring import ConstBitStream, Bits
from .field_codec import FieldCodec
from math import ceil, log2


class IeeeFloatCodec(FieldCodec):
    def __init__(self, num_bits=32, **kwargs):
        if num_bits not in (32, 64):
            raise ValueError("Only 32 or 64 bit widths are supported for IEEE floating point codecs")
        self.num_bits = num_bits

    def encode(self, value: float):
        value = float(value)
        encoded_bits = Bits(float=value, length=self.num_bits)
        encoded_value = encoded_bits.float
        ## print("encode varint {} bits as {}".format(num_bits, encoded_bits))
        return encoded_bits, encoded_value

    def decode(self, bits_to_decode: ConstBitStream):
        value = bits_to_decode.read('floatbe:{}'.format(self.num_bits))
        return value

    @property
    def max_length_bits(self):
        return self.num_bits

    @property
    def min_length_bits(self):
        return self.num_bits


class IeeeFloat32Codec(IeeeFloatCodec):
    def __init__(self, **kwargs):
        super(IeeeFloat32Codec, self).__init__(num_bits=32)


class IeeeFloat64Codec(IeeeFloatCodec):
    def __init__(self, **kwargs):
        super(IeeeFloat64Codec, self).__init__(num_bits=64)


from bitstring import ConstBitStream, Bits
from .field_codec import FieldCodec
from math import ceil, log2
from rospy import Time


class RosTimeCodec(FieldCodec):
    def __init__(self, precision: int = 0, epoch_start=1622520000, epoch_end=(2**31 - 1), **kwargs):
        self.max_value = epoch_end
        self.min_value = epoch_start
        self.precision = int(precision)

        self.value_range = int(round(self.max_value - self.min_value, precision) * 10 ** precision)
        self.num_bits = ceil(log2(self.value_range))
        #print("Float codec: precision {} num_bits {}".format(precision, self.num_bits))

    def encode(self, value: Time):
        value = value.to_sec()
        if value < self.min_value:
            value = self.min_value
        elif value > self.max_value:
            value = self.max_value

        encoded_value = int(round(value - self.min_value, self.precision) * 10 ** self.precision)
        #print(value, self.min_value, self.precision)
        rounded_value = round(value - self.min_value, self.precision) + self.min_value  # Used only for validation
        encoded_bits = Bits(uint=encoded_value, length=self.num_bits)

        return encoded_bits, rounded_value

    def decode(self, bits_to_decode: ConstBitStream):
        float_offset = bits_to_decode.read('uint:{}'.format(self.num_bits)) / 10 ** self.precision
        value = Time.from_sec(self.min_value + float_offset)
        return value

    @property
    def max_length_bits(self):
        return self.num_bits

    @property
    def min_length_bits(self):
        return self.num_bits


from bitstring import ConstBitStream, Bits
from .field_codec import FieldCodec
from math import ceil, log2


class LinspaceFloatCodec(FieldCodec):
    def __init__(self, min_value: float, max_value: float,
                 resolution: float = None, num_values: int = None, num_bits: int = None, **kwargs):
        self.max_value = float(max_value)
        self.min_value = float(min_value)
        self.value_range = max_value - min_value
        if num_values or num_bits:
            if resolution:
                raise ValueError("LinspaceFloatCodec supports setting only one of num_values, num_bits, or resolution.")
            if num_bits:
                if num_values:
                    raise ValueError("LinspaceFloatCodec supports setting either num_values or num_bits, not both.")
                num_values = 2 ** num_bits
            resolution = self.value_range / num_values

        self.resolution = float(resolution)

        self.num_values = self.value_range // self.resolution
        # if the max value isn't an integer multiple of the resolution from the min value, it won't be encoded.
        self.num_bits = ceil(log2(self.num_values))

    def encode(self, value: float):
        value = float(value)
        if value < self.min_value:
            value = self.min_value
        elif value > self.max_value:
            value = self.max_value
        offset = value - self.min_value
        discretized_offset = int(offset // self.resolution)
        encoded_value = self.min_value + (discretized_offset * self.resolution)
        encoded_bits = Bits(uint=discretized_offset, length=self.num_bits)
        ## print("encode varint {} bits as {}".format(num_bits, encoded_bits))
        return encoded_bits, encoded_value

    def decode(self, bits_to_decode: ConstBitStream):
        discretized_offset = bits_to_decode.read('uint:{}'.format(self.num_bits))
        value = self.min_value + (discretized_offset * self.resolution)
        return value

    @property
    def max_length_bits(self):
        return self.num_bits

    @property
    def min_length_bits(self):
        return self.num_bits


from bitstring import BitArray, ConstBitStream
from .field_codec import FieldCodec


class PaddingCodec(FieldCodec):
    def __init__(self, num_bits, **kwargs):
        self.num_bits = int(num_bits)
        pass

    def encode(self, value=None):
        value_bits = BitArray(uint=0, length=self.num_bits)
        return value_bits, None

    def decode(self, bits_to_decode: ConstBitStream):
        bits_to_decode.read('pad:{}'.format(self.num_bits))
        return None

    def min_length_bits(self):
        return self.num_bits

    def max_length_bits(self):
        return self.num_bits